#include <fstream>
#include <iostream>
#include <cmath>

#define M1 10
#define NTST 20
#define DT 0.01
#define DY 1.0
#define DZ 1.0
#define ALPHA 1.0
using namespace std;

void init_mesh(double* dx, double* V, double* Af)
{
    for(int i=1;i<=M1+1;i++) dx[i]=1.0/double(M1);
    dx[1]=0.5*dx[1];
    dx[M1+1]=0.5*dx[M1+1];
    for(int i=1;i<=M1;i++)
    {
        Af[i]=DY*DZ;
        V[i]=Af[i]*1.0/double(M1);
    }
    Af[M1+1]=DY*DZ;
}
void init_phi(double* phi)
{
    for(int i=1;i<=M1;i++) phi[i]=0.0;
}
void bcond(double* phi)
{
    phi[0]=-100.0;
//    phi[M1+1]=200.0;	// for Dirichlet at i=M1
    phi[M1+1]=phi[M1];	// for Neumann at i=M1
}
void calc_source(double* phi, double* source, double* dx, double* V, double* Af)
{
    for(int i=1;i<=M1;i++)
    {
        source[i]=1.0/DT*phi[i]*V[i];
    }
    source[1]+=ALPHA*Af[1]*phi[0]/dx[1];
}
void calc_coef(double* coef, double* source, double* dx, double* V, double* Af, const int& nTime)
{
    double coefRef[M1+1][M1+1];
    for(int i=1;i<=M1;i++) for(int j=1;j<=M1;j++) coefRef[i][j]=0.0;
    for(int i=1;i<=M1;i++)
    {
        coefRef[i][i]=1.0*V[i]/DT+ALPHA*(Af[i+1]/dx[i+1]+Af[i]/dx[i]);
        if(i!=1) coefRef[i-1][i]=-ALPHA*Af[i]/dx[i];
        if(i!=M1) coefRef[i+1][i]=-ALPHA*Af[i+1]/dx[i+1];
    }
    coefRef[M1][M1]=1.0*V[M1]/DT+ALPHA*Af[M1]/dx[M1];	// for Neumann at i=M1

    if(nTime==NTST-1)
    {
        ofstream coefFile("MATRIX_COEF.dat");
        for(int j=1;j<=M1;j++)
        {
            for(int i=1;i<=M1;i++) coefFile << coefRef[i][j] << " ";
            coefFile << " | " << source[j] << endl;
        }
        coefFile.close();
    }
    for(int i=1;i<=M1;i++) for(int j=1;j<=M1;j++) coef[i*M1+j]=coefRef[i][j];
}
void solve(double* coef, double* phi, double* source, const int& nTime)
{
    int nIter=1;
    double coefRef[M1+1][M1+1], oldPhi[M1+2], residual;

    for(int i=1;i<=M1;i++) for(int j=1;j<=M1;j++) coefRef[i][j]=coef[i*M1+j];
    do
    {
        for(int i=1;i<=M1;i++) oldPhi[i]=phi[i];
        for(int i=1;i<=M1;i++)
        {
            double sumLHS=0.0;
            for(int iLHS=1;iLHS<=M1;iLHS++) sumLHS+=coefRef[iLHS][i]*oldPhi[iLHS];
            sumLHS-=coefRef[i][i]*oldPhi[i];
            phi[i]=(source[i]-sumLHS)/coefRef[i][i];
        }
        residual=0.0;
        for(int i=1;i<=M1;i++) residual+=fabs(phi[i]-oldPhi[i])/double(M1);
        cout << nTime << " : RES = " << residual << ", ITER = " << nIter << endl;
        nIter++;
    } while(residual>0.0001);
}
void write_phi(double* phi)
{
    ofstream os("phi");
    for(int i=1;i<=M1;i++) os << phi[i] << endl;
    os.close();
}
int main()
{
    double phi[M1+2], source[M1+1], coef[M1+1][M1+1];
    double dx[M1+2],V[M1+1], Af[M1+2];
    init_mesh(dx,V,Af);
    init_phi(phi);
    for(int nTime=0;nTime<NTST;nTime++)
    {
        bcond(phi);
        calc_source(phi,source,dx,V,Af);
        calc_coef((double*)coef,source,dx,V,Af,nTime);
        solve((double*)coef,phi,source,nTime);
    }
    write_phi(phi);
    return 0;
}
