#include <fstream>
#include <iostream>
#include <cmath>

#define NTST 20
#define DT 0.01
#define DY 1.0
#define DZ 1.0
#define ALPHA 1.0
using namespace std;

#include "List.H"

void init_mesh(List<double>& dx, List<double>& V, List<double>& Af)
{
    for(int i=1;i<dx.size();i++) dx[i]=1.0/double(V.size()-1);
    dx[1]=0.5*dx[1];
    dx[dx.size()-1]=0.5*dx[dx.size()-1];
    for(int i=1;i<V.size();i++)
    {
        Af[i]=DY*DZ;
        V[i]=Af[i]*1.0/double(V.size()-1);
    }
    Af[Af.size()-1]=DY*DZ;
}
void init_phi(List<double>& phi)
{
    for(int i=1;i<phi.size()-1;i++) phi[i]=0.0;
}
void bcond(List<double>& phi)
{
    phi[0]=-100.0;
//    phi[phi.size()-1]=200.0;	// for Dirichlet at i=M1
    phi[phi.size()-1]=phi[phi.size()-2];	// for Neumann at i=M1
}
void calc_source(List<double> phi, List<double>& source, List<double>& dx, List<double>& V, List<double>& Af)
{
    for(int i=1;i<source.size();i++)
    {
        source[i]=1.0/DT*phi[i]*V[i];
    }
    source[1]+=ALPHA*Af[1]*phi[0]/dx[1];
}
void calc_coef(List<List<double> >& coef, List<double> source, List<double>& dx, List<double>& V, List<double>& Af, const int& nTime)
{
    for(int i=1;i<coef.size();i++) for(int j=1;j<coef[i].size();j++) coef[i][j]=0.0;
    for(int i=1;i<coef.size();i++)
    {
        coef[i][i]=1.0*V[i]/DT+ALPHA*(Af[i+1]/dx[i+1]+Af[i]/dx[i]);
        if(i!=1) coef[i-1][i]=-ALPHA*Af[i]/dx[i];
        if(i!=coef.size()-1) coef[i+1][i]=-ALPHA*Af[i+1]/dx[i+1];
    }
    coef[coef.size()-1][coef.size()-1]=1.0*V[coef.size()-1]/DT+ALPHA*Af[coef.size()-1]/dx[coef.size()-1];   // for Neumann at i=M1

    if(nTime==NTST-1)
    {
        ofstream coefFile("MATRIX_COEF.dat");
        for(int j=1;j<coef.size();j++)
        {
            for(int i=1;i<coef.size();i++) coefFile << coef[i][j] << " ";
            coefFile << " | " << source[j] << endl;
        }
        coefFile.close();
    }
}
void solve(List<List<double> > coef, List<double>& phi, List<double> source, const int& nTime)
{
    int nIter=1;
    double residual;
    List<double> oldPhi(phi.size());
    do
    {
        for(int i=1;i<phi.size()-1;i++) oldPhi[i]=phi[i];
        for(int i=1;i<phi.size()-1;i++)
        {
            double sumLHS=0.0;
            for(int iLHS=1;iLHS<oldPhi.size()-1;iLHS++) sumLHS+=coef[iLHS][i]*oldPhi[iLHS];
            sumLHS-=coef[i][i]*oldPhi[i];
            phi[i]=(source[i]-sumLHS)/coef[i][i];
        }
        residual=0.0;
        for(int i=1;i<phi.size()-1;i++) residual+=fabs(phi[i]-oldPhi[i])/double(phi.size()-2);
        cout << nTime << " : RES = " << residual << ", ITER = " << nIter << endl;
        nIter++;
    } while(residual>0.0001);
}
void write_phi(List<double> phi)
{
    ofstream os("phi");
    for(int i=1;i<phi.size()-1;i++) os << phi[i] << endl;
    os.close();
}
int main()
{
    int M1=10;
    List<double> phi(M1+2), source(M1+1), dx(M1+2), V(M1+1), Af(M1+2);
    List<List<double> > coef(M1+1);
    for(int i=1;i<coef.size();i++) coef[i].setSize(M1+1);
    init_mesh(dx,V,Af);
    init_phi(phi);
    for(int nTime=0;nTime<NTST;nTime++)
    {
        bcond(phi);
        calc_source(phi,source,dx,V,Af);
        calc_coef(coef,source,dx,V,Af,nTime);
        solve(coef,phi,source,nTime);
    }
    write_phi(phi);
    return 0;
}
